=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2008-05-22 05:58:50 +0000
+++ bzrlib/remote.py	2008-05-22 07:49:06 +0000
@@ -1515,6 +1515,23 @@
             return []
         return result
 
+    def _set_last_revision_descendant(self, revision_id, other_branch):
+        path = self.bzrdir._path_for_remote_call(self._client)
+        try:
+            response = self._client.call('Branch.set_last_revision_descendant',
+                path, self._lock_token, self._repo_lock_token, revision_id)
+        except errors.ErrorFromSmartServer, err:
+            if err.error_verb == 'NoSuchRevision':
+                raise NoSuchRevision(self, revision_id)
+            elif err.error_verb == 'NotDescendant':
+                raise errors.DivergedBranches(self, other_branch)
+            raise
+        self._clear_cached_state()
+        if len(response) != 2 and response[0] != 'ok':
+            raise errors.UnexpectedSmartServerResponse(response)
+        new_revno = response[1]
+        self._last_revision_info_cache = new_revno, revision_id
+
     def _set_last_revision(self, revision_id):
         path = self.bzrdir._path_for_remote_call(self._client)
         self._clear_cached_state()
@@ -1608,9 +1625,16 @@
 
     def generate_revision_history(self, revision_id, last_rev=None,
                                   other_branch=None):
+        medium = self._client._medium
+        if medium._remote_is_at_least_1_6:
+            try:
+                self._set_last_revision_descendant(revision_id, other_branch)
+                return
+            except UnknownSmartMethod:
+                medium._remote_is_at_least_1_6 = False
         self._clear_cached_state()
         self._ensure_real()
-        return self._real_branch.generate_revision_history(
+        self._real_branch.generate_revision_history(
             revision_id, last_rev=last_rev, other_branch=other_branch)
 
     @property
@@ -1622,52 +1646,32 @@
         self._ensure_real()
         return self._real_branch.set_push_location(location)
 
+    @needs_write_lock
     def update_revisions(self, other, stop_revision=None, overwrite=False):
-        if overwrite:
-            self._clear_cached_state()
-            self._ensure_real()
-            return self._real_branch.update_revisions(
-                other, stop_revision=stop_revision, overwrite=True)
-        # XXX: this code is substantially copy-and-pasted from
-        # Branch.update_revisions.  This is however much faster than calling
-        # the same code on _real_branch, because it will use RPCs and cache
-        # results.
         other.lock_read()
         try:
-            other_last_revno, other_last_revision = other.last_revision_info()
             if stop_revision is None:
-                stop_revision = other_last_revision
+                stop_revision = other.last_revision()
                 if revision.is_null(stop_revision):
                     # if there are no commits, we're done.
                     return
             # whats the current last revision, before we fetch [and change it
             # possibly]
-            last_rev = revision.ensure_null(self.last_revision())
-            # we fetch here so that we don't process data twice in the common
-            # case of having something to pull, and so that the check for 
-            # already merged can operate on the just fetched graph, which will
-            # be cached in memory.
             self.fetch(other, stop_revision)
-            # Check to see if one is an ancestor of the other
-            heads = self.repository.get_graph().heads([stop_revision,
-                                                       last_rev])
-            if heads == set([last_rev]):
-                # The current revision is a decendent of the target,
-                # nothing to do
-                return
-            elif heads == set([stop_revision, last_rev]):
-                # These branches have diverged
-                raise errors.DivergedBranches(self, other)
-            elif heads != set([stop_revision]):
-                raise AssertionError("invalid heads: %r" % heads)
-            if other_last_revision == stop_revision:
-                self.set_last_revision_info(other_last_revno,
-                                            other_last_revision)
-            else:
-                # XXX: In Branch.update_revisions this code is more
-                # complicated.  Here we just allow the remote side to generate
-                # the new history for us.
+
+            if overwrite:
                 self._set_last_revision(stop_revision)
+            else:
+                medium = self._client._medium
+                if medium._remote_is_at_least_1_6:
+                    try:
+                        self._set_last_revision_descendant(stop_revision, other)
+                        return
+                    except UnknownSmartMethod:
+                        medium._remote_is_at_least_1_6 = False
+                last_rev = revision.ensure_null(self.last_revision())
+                self.generate_revision_history(
+                    stop_revision, last_rev=last_rev, other_branch=other)
         finally:
             other.unlock()
 

=== modified file 'bzrlib/smart/branch.py'
--- bzrlib/smart/branch.py	2008-05-07 06:03:09 +0000
+++ bzrlib/smart/branch.py	2008-05-22 07:49:06 +0000
@@ -128,6 +128,21 @@
         return SuccessfulSmartServerResponse(('ok',))
 
 
+class SmartServerBranchRequestSetLastRevisionDescendant(SmartServerLockedBranchRequest):
+    """New in 1.6."""
+    
+    def do_with_locked_branch(self, branch, new_last_revision_id):
+        try:
+            last_rev = branch.last_revision()
+            branch.generate_revision_history(new_last_revision_id, last_rev)
+        except errors.DivergedBranches:
+            return FailedSmartServerResponse(('NotDescendant',))
+        except errors.NoSuchRevision:
+            return FailedSmartServerResponse(
+                ('NoSuchRevision', new_last_revision_id))
+        return SuccessfulSmartServerResponse(('ok', branch.revno()))
+
+
 class SmartServerBranchRequestSetLastRevisionInfo(
     SmartServerLockedBranchRequest):
     """Branch.set_last_revision_info.  Sets the revno and the revision ID of

=== modified file 'bzrlib/smart/medium.py'
--- bzrlib/smart/medium.py	2008-05-20 00:42:19 +0000
+++ bzrlib/smart/medium.py	2008-05-22 07:49:06 +0000
@@ -509,6 +509,7 @@
         # requests until we get an error saying otherwise.  (1.2 adds some
         # requests that send bodies, which confuses older servers.)
         self._remote_is_at_least_1_2 = True
+        self._remote_is_at_least_1_6 = True
 
     def accept_bytes(self, bytes):
         self._accept_bytes(bytes)

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2008-05-21 11:53:44 +0000
+++ bzrlib/smart/request.py	2008-05-22 07:49:06 +0000
@@ -386,6 +386,9 @@
     'Branch.set_last_revision_info', 'bzrlib.smart.branch',
     'SmartServerBranchRequestSetLastRevisionInfo')
 request_handlers.register_lazy(
+    'Branch.set_last_revision_descendant', 'bzrlib.smart.branch',
+    'SmartServerBranchRequestSetLastRevisionDescendant')
+request_handlers.register_lazy(
     'Branch.unlock', 'bzrlib.smart.branch', 'SmartServerBranchRequestUnlock')
 request_handlers.register_lazy(
     'BzrDir.find_repository', 'bzrlib.smart.bzrdir', 'SmartServerRequestFindRepositoryV1')

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2008-05-22 05:20:46 +0000
+++ bzrlib/tests/test_smart.py	2008-05-22 07:49:06 +0000
@@ -397,26 +397,14 @@
             request.execute(''))
 
 
-class TestSmartServerBranchRequestSetLastRevision(tests.TestCaseWithMemoryTransport):
+class SetLastRevisionCommonTests(object):
+    """Mixin class for tests common to Branch.set_last_revision* variants."""
 
-    def test_empty(self):
-        backing = self.get_transport()
-        request = smart.branch.SmartServerBranchRequestSetLastRevision(backing)
-        b = self.make_branch('.')
-        branch_token = b.lock_write()
-        repo_token = b.repository.lock_write()
-        b.repository.unlock()
-        try:
-            self.assertEqual(SmartServerResponse(('ok',)),
-                request.execute(
-                    '', branch_token, repo_token,
-                    'null:'))
-        finally:
-            b.unlock()
+    request_class = None
 
     def test_not_present_revision_id(self):
         backing = self.get_transport()
-        request = smart.branch.SmartServerBranchRequestSetLastRevision(backing)
+        request = self.request_class(backing)
         b = self.make_branch('.')
         branch_token = b.lock_write()
         repo_token = b.repository.lock_write()
@@ -431,50 +419,170 @@
         finally:
             b.unlock()
 
-    def test_revision_id_present(self):
-        backing = self.get_transport()
-        request = smart.branch.SmartServerBranchRequestSetLastRevision(backing)
-        tree = self.make_branch_and_memory_tree('.')
-        tree.lock_write()
-        tree.add('')
-        rev_id_utf8 = u'\xc8'.encode('utf-8')
-        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
-        r2 = tree.commit('2nd commit')
-        tree.unlock()
-        branch_token = tree.branch.lock_write()
-        repo_token = tree.branch.repository.lock_write()
-        tree.branch.repository.unlock()
-        try:
-            self.assertEqual(
-                SmartServerResponse(('ok',)),
-                request.execute(
-                    '', branch_token, repo_token,
-                    rev_id_utf8))
-            self.assertEqual([rev_id_utf8], tree.branch.revision_history())
-        finally:
-            tree.branch.unlock()
-
-    def test_revision_id_present2(self):
-        backing = self.get_transport()
-        request = smart.branch.SmartServerBranchRequestSetLastRevision(backing)
-        tree = self.make_branch_and_memory_tree('.')
-        tree.lock_write()
-        tree.add('')
-        rev_id_utf8 = u'\xc8'.encode('utf-8')
-        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
-        r2 = tree.commit('2nd commit')
-        tree.unlock()
-        tree.branch.set_revision_history([])
-        branch_token = tree.branch.lock_write()
-        repo_token = tree.branch.repository.lock_write()
-        tree.branch.repository.unlock()
-        try:
-            self.assertEqual(
-                SmartServerResponse(('ok',)),
-                request.execute(
-                    '', branch_token, repo_token,
-                    rev_id_utf8))
-            self.assertEqual([rev_id_utf8], tree.branch.revision_history())
+
+class TestSmartServerBranchRequestSetLastRevision(
+    tests.TestCaseWithMemoryTransport, SetLastRevisionCommonTests):
+
+    request_class = smart.branch.SmartServerBranchRequestSetLastRevision
+
+    def test_revision_id_present2(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        tree = self.make_branch_and_memory_tree('.')
+        tree.lock_write()
+        tree.add('')
+        rev_id_utf8 = u'\xc8'.encode('utf-8')
+        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
+        r2 = tree.commit('2nd commit')
+        tree.unlock()
+        tree.branch.set_revision_history([])
+        branch_token = tree.branch.lock_write()
+        repo_token = tree.branch.repository.lock_write()
+        tree.branch.repository.unlock()
+        try:
+            self.assertEqual(
+                SuccessfulSmartServerResponse(('ok',)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    rev_id_utf8))
+            self.assertEqual([rev_id_utf8], tree.branch.revision_history())
+        finally:
+            tree.branch.unlock()
+
+
+    def test_revision_id_previous(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        tree = self.make_branch_and_memory_tree('.')
+        tree.lock_write()
+        tree.add('')
+        rev_id_utf8 = u'\xc8'.encode('utf-8')
+        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
+        r2 = tree.commit('2nd commit')
+        tree.unlock()
+        branch_token = tree.branch.lock_write()
+        repo_token = tree.branch.repository.lock_write()
+        tree.branch.repository.unlock()
+        try:
+            self.assertEqual(
+                SuccessfulSmartServerResponse(('ok',)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    rev_id_utf8))
+            self.assertEqual([rev_id_utf8], tree.branch.revision_history())
+        finally:
+            tree.branch.unlock()
+
+
+class TestSmartServerBranchRequestSetLastRevisionDescendant(
+    tests.TestCaseWithMemoryTransport, SetLastRevisionCommonTests):
+
+    request_class = \
+        smart.branch.SmartServerBranchRequestSetLastRevisionDescendant
+
+    def test_empty(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        b = self.make_branch('.')
+        branch_token = b.lock_write()
+        repo_token = b.repository.lock_write()
+        b.repository.unlock()
+        try:
+            self.assertEqual(SmartServerResponse(('ok',)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    'null:'))
+        finally:
+            b.unlock()
+
+    def test_empty(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        b = self.make_branch('.')
+        branch_token = b.lock_write()
+        repo_token = b.repository.lock_write()
+        b.repository.unlock()
+        try:
+            self.assertEqual(SmartServerResponse(('ok', 0)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    'null:'))
+        finally:
+            b.unlock()
+
+    def test_revision_id_present2(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        tree = self.make_branch_and_memory_tree('.')
+        tree.lock_write()
+        tree.add('')
+        rev_id_utf8 = u'\xc8'.encode('utf-8')
+        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
+        r2 = tree.commit('2nd commit')
+        tree.unlock()
+        tree.branch.set_revision_history([])
+        branch_token = tree.branch.lock_write()
+        repo_token = tree.branch.repository.lock_write()
+        tree.branch.repository.unlock()
+        try:
+            self.assertEqual(
+                SuccessfulSmartServerResponse(('ok', 1)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    rev_id_utf8))
+            self.assertEqual([rev_id_utf8], tree.branch.revision_history())
+        finally:
+            tree.branch.unlock()
+
+    def test_revision_id_previous(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        tree = self.make_branch_and_memory_tree('.')
+        tree.lock_write()
+        tree.add('')
+        rev_id_utf8 = u'\xc8'.encode('utf-8')
+        r1 = tree.commit('1st commit', rev_id=rev_id_utf8)
+        r2 = tree.commit('2nd commit', rev_id='second-rev')
+        tree.unlock()
+        branch_token = tree.branch.lock_write()
+        repo_token = tree.branch.repository.lock_write()
+        tree.branch.repository.unlock()
+        try:
+            self.assertEqual(
+                FailedSmartServerResponse(('NotDescendant',)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    rev_id_utf8))
+            self.assertEqual('second-rev', tree.branch.last_revision())
+        finally:
+            tree.branch.unlock()
+
+    def test_branch_diverged(self):
+        backing = self.get_transport()
+        request = self.request_class(backing)
+        tree = self.make_branch_and_memory_tree('.')
+        tree.lock_write()
+        tree.add('')
+        r1 = tree.commit('1st commit')
+        revno_1, revid_1 = tree.branch.last_revision_info()
+        r2 = tree.commit('2nd commit', rev_id='child-1')
+        # Undo the second commit
+        tree.branch.set_last_revision_info(revno_1, revid_1)
+        tree.set_parent_ids([revid_1])
+        # Make a new second commit, child-2.  child-2 has diverged from
+        # child-1.
+        new_r2 = tree.commit('2nd commit', rev_id='child-2')
+        tree.unlock()
+        branch_token = tree.branch.lock_write()
+        repo_token = tree.branch.repository.lock_write()
+        tree.branch.repository.unlock()
+        try:
+            self.assertEqual(
+                FailedSmartServerResponse(('NotDescendant',)),
+                request.execute(
+                    '', branch_token, repo_token,
+                    'child-1'))
+            self.assertEqual('child-2', tree.branch.last_revision())
         finally:
             tree.branch.unlock()
 

